@@ -29,6 +29,7 @@ from gettext import gettext as _
 from os.path import commonprefix
 
 from gi.repository import Gdk
+from gi.repository import Gio
 from gi.repository import GLib
 from gi.repository import GObject
 from gi.repository import Gtk
@@ -36,7 +37,6 @@ from gi.repository import Pango
 
 from pynicotine import slskmessages
 from pynicotine.gtkgui.dialogs import entry_dialog
-from pynicotine.gtkgui.nowplaying import NowPlaying
 from pynicotine.gtkgui.roomwall import RoomWall
 from pynicotine.gtkgui.roomwall import Tickers
 from pynicotine.gtkgui.utils import append_line
@@ -45,6 +45,7 @@ from pynicotine.gtkgui.utils import humanize
 from pynicotine.gtkgui.utils import human_speed
 from pynicotine.gtkgui.utils import IconNotebook
 from pynicotine.gtkgui.utils import initialise_columns
+from pynicotine.gtkgui.utils import load_ui_elements
 from pynicotine.gtkgui.utils import PopupMenu
 from pynicotine.gtkgui.utils import press_header
 from pynicotine.gtkgui.utils import scroll_bottom
@@ -52,6 +53,9 @@ from pynicotine.gtkgui.utils import TextSearchBar
 from pynicotine.gtkgui.utils import expand_alias
 from pynicotine.gtkgui.utils import is_alias
 from pynicotine.gtkgui.utils import show_country_tooltip
+from pynicotine.gtkgui.utils import set_widget_color
+from pynicotine.gtkgui.utils import set_widget_font
+from pynicotine.gtkgui.utils import update_widget_visuals
 from pynicotine.logfacility import log
 from pynicotine.utils import clean_file
 from pynicotine.utils import cmp
@@ -144,7 +148,8 @@ class RoomsControl:
         self.chat_notebook.notebook.connect("switch-page", self.on_switch_page)
         self.chat_notebook.notebook.connect("page-reordered", self.on_reordered_page)
 
-        self.frame.set_text_bg(self.frame.roomlist.SearchRooms)
+        self.frame.roomlist.update_visuals()
+        self.update_visuals()
 
     def is_private_room_owned(self, room):
 
@@ -199,8 +204,6 @@ class RoomsControl:
             cellrenderer.set_property("weight", Pango.Weight.NORMAL)
             cellrenderer.set_property("underline", Pango.Underline.NONE)
 
-        self.frame.cell_data_func(column, cellrenderer, model, iterator)
-
     def on_reordered_page(self, notebook, page, page_num, force=0):
 
         room_tab_order = {}
@@ -230,9 +233,9 @@ class RoomsControl:
         # Save
         self.frame.np.config.sections["server"]["autojoin"] = new_autojoin
 
-    def on_switch_page(self, notebook, page, page_num, force=0):
+    def on_switch_page(self, notebook, page, page_num, forceupdate=False):
 
-        if self.frame.MainNotebook.get_current_page() != self.frame.MainNotebook.page_num(self.frame.chathbox) and not force:
+        if self.frame.MainNotebook.get_current_page() != self.frame.MainNotebook.page_num(self.frame.chathbox) and not forceupdate:
             return
 
         for name, room in self.joinedrooms.items():
@@ -672,12 +675,11 @@ class RoomsControl:
 
         room.say_chat_room(msg, text, public=True)
 
-    def update_colours(self):
-
-        self.frame.set_text_bg(self.frame.roomlist.SearchRooms)
+    def update_visuals(self):
 
         for room in self.joinedrooms.values():
-            room.change_colours()
+            room.update_visuals()
+            room.update_tags()
 
     def save_columns(self):
 
@@ -745,27 +747,10 @@ class ChatRoom:
     def __init__(self, roomsctrl, room, users, meta=False):
 
         self.roomsctrl = roomsctrl
-
         self.frame = roomsctrl.frame
 
         # Build the window
-        builder = Gtk.Builder()
-
-        builder.set_translation_domain('nicotine')
-        builder.add_from_file(os.path.join(os.path.dirname(os.path.realpath(__file__)), "ui", "chatrooms.ui"))
-
-        self.chat_room_tab = builder.get_object("ChatRoomTab")
-
-        for i in builder.get_objects():
-            try:
-                self.__dict__[Gtk.Buildable.get_name(i)] = i
-            except TypeError:
-                pass
-
-        self.chat_room_tab.remove(self.Main)
-        self.chat_room_tab.destroy()
-
-        builder.connect_signals(self)
+        load_ui_elements(self, os.path.join(self.frame.gui_dir, "ui", "chatrooms.ui"))
 
         self.tickers = Tickers()
         self.room_wall = RoomWall(self.frame, self)
@@ -786,6 +771,9 @@ class ChatRoom:
         TextSearchBar(self.ChatScroll, self.ChatSearchBar, self.ChatSearchEntry)
 
         # Spell Check
+        if self.frame.spell_checker is None:
+            self.frame.init_spell_checker()
+
         if self.frame.spell_checker and self.frame.np.config.sections["ui"]["spellcheck"]:
             from gi.repository import Gspell
             spell_buffer = Gspell.EntryBuffer.get_from_gtk_entry_buffer(self.ChatEntry.get_buffer())
@@ -812,7 +800,7 @@ class ChatRoom:
             self.AutoJoin.set_active(True)
 
         if room not in config["columns"]["chatrooms_widths"]:
-            config["columns"]["chatrooms_widths"][room] = [0, 25, 100, 0, 0]
+            config["columns"]["chatrooms_widths"][room] = [25, 25, 100, 0, 0]
 
         widths = self.frame.np.config.sections["columns"]["chatrooms_widths"][room]
         self.cols = cols = initialise_columns(
@@ -820,8 +808,8 @@ class ChatRoom:
             [_("Status"), widths[0], "pixbuf"],
             [_("Country"), widths[1], "pixbuf"],
             [_("User"), widths[2], "text", self.user_column_draw],
-            [_("Speed"), widths[3], "number", self.frame.cell_data_func],
-            [_("Files"), widths[4], "number", self.frame.cell_data_func]
+            [_("Speed"), widths[3], "number"],
+            [_("Files"), widths[4], "number"]
         )
 
         cols[0].set_sort_column_id(5)
@@ -879,7 +867,8 @@ class ChatRoom:
 
         self.usersmodel.set_sort_column_id(2, Gtk.SortType.ASCENDING)
 
-        self.update_colours()
+        self.create_tags()
+        self.update_visuals()
 
         self.UserList.set_model(self.usersmodel)
 
@@ -1042,8 +1031,8 @@ class ChatRoom:
         # Double click starts a private message
         if event.button != 3:
             if event.type == Gdk.EventType._2BUTTON_PRESS:
-                self.frame.privatechats.send_message(user, None, 1)
-                self.frame.change_main_page(None, "private")
+                self.frame.privatechats.send_message(user, show_user=True)
+                self.frame.change_main_page("private")
             return
 
         self.popup_menu.editing = True
@@ -1051,7 +1040,7 @@ class ChatRoom:
 
         me = (self.popup_menu.user is None or self.popup_menu.user == self.frame.np.config.sections["server"]["login"])
 
-        self.menu_add_to_list.set_active(user in [i[0] for i in self.frame.np.config.sections["server"]["userlist"]])
+        self.menu_add_to_list.set_active(user in (i[0] for i in self.frame.np.config.sections["server"]["userlist"]))
         self.menu_ban_user.set_active(user in self.frame.np.config.sections["server"]["banlist"])
         self.menu_ignore_user.set_active(user in self.frame.np.config.sections["server"]["ignorelist"])
         self.menu_block_user.set_active(self.frame.user_ip_is_blocked(user))
@@ -1135,19 +1124,18 @@ class ChatRoom:
             if tag == self.tag_hilite:
 
                 self.roomsctrl.chat_notebook.request_hilite(self.Main)
-                self.frame.chat_request_icon(1, self.Main)
+                self.frame.request_tab_icon(self.frame.ChatTabLabel, status=1)
 
                 if self.frame.np.config.sections["notifications"]["notification_popup_chatroom_mention"]:
                     self.frame.notifications.new_notification(
                         text,
                         title=_("%s mentioned you in the %s room") % (user, self.room),
-                        soundnamenotify="bell-window-system",
-                        soundnamewin="SystemExclamation"
+                        priority=Gio.NotificationPriority.HIGH
                     )
 
             else:
                 self.roomsctrl.chat_notebook.request_changed(self.Main)
-                self.frame.chat_request_icon(0)
+                self.frame.request_tab_icon(self.frame.ChatTabLabel, status=0)
 
             if self.roomsctrl.chat_notebook.get_current_page() != self.roomsctrl.chat_notebook.page_num(self.roomsctrl.joinedrooms[self.room].Main) or \
                 self.frame.MainNotebook.get_current_page() != self.frame.MainNotebook.page_num(self.frame.chathbox) or \
@@ -1161,7 +1149,8 @@ class ChatRoom:
                 elif self.frame.np.config.sections["notifications"]["notification_popup_chatroom"]:
                     self.frame.notifications.new_notification(
                         text,
-                        title=_("Message by %s in the %s room") % (user, self.room)
+                        title=_("Message by %s in the %s room") % (user, self.room),
+                        priority=Gio.NotificationPriority.HIGH
                     )
 
         if text[:4] == "/me ":
@@ -1221,9 +1210,9 @@ class ChatRoom:
             color = self.get_user_status_color(self.usersmodel.get_value(self.users[user], 5))
 
         if user not in self.tag_users:
-            self.tag_users[user] = self.makecolour(self.ChatScroll.get_buffer(), color, user)
+            self.tag_users[user] = self.create_tag(self.ChatScroll.get_buffer(), color, user)
         else:
-            self.changecolour(self.tag_users[user], color)
+            self.update_tag_visuals(self.tag_users[user], color)
 
     def thread_alias(self, alias):
 
@@ -1286,14 +1275,13 @@ class ChatRoom:
         elif cmd == "/ip":
             if byteargs:
                 user = byteargs
-                if user not in self.frame.np.ip_requested:
-                    self.frame.np.ip_requested.append(user)
+                self.frame.np.ip_requested.add(user)
                 self.frame.np.queue.put(slskmessages.GetPeerAddress(user))
 
         elif cmd == "/pm":
             if byteargs:
-                self.frame.privatechats.send_message(byteargs, None, 1)
-                self.frame.on_private_chat(None)
+                self.frame.privatechats.send_message(byteargs, show_user=True)
+                self.frame.change_main_page("private")
 
         elif cmd in ["/m", "/msg"]:
             if byteargs:
@@ -1370,7 +1358,7 @@ class ChatRoom:
             self.frame.on_away(None)
 
         elif cmd in ["/q", "/quit", "/exit"]:
-            self.frame.on_exit(None)
+            self.frame.on_quit(None)
             return  # Avoid gsignal warning
 
         elif cmd == "/now":
@@ -1426,10 +1414,7 @@ class ChatRoom:
         self.frame.np.queue.put(slskmessages.SayChatroom(self.room, text))
 
     def display_now_playing(self):
-        if self.frame.now_playing is None:
-            self.frame.now_playing = NowPlaying(self.frame)
-
-        self.frame.now_playing.display_now_playing(None, test=0, callback=self.say)
+        self.frame.now_playing.display_now_playing(callback=self.say)
 
     def user_joined_room(self, username, userdata):
 
@@ -1474,7 +1459,7 @@ class ChatRoom:
         # Remove from completion list, and completion drop-down
         if self.frame.np.config.sections["words"]["tab"]:
 
-            if username in self.clist and username not in [i[0] for i in self.frame.np.config.sections["server"]["userlist"]]:
+            if username in self.clist and username not in (i[0] for i in self.frame.np.config.sections["server"]["userlist"]):
 
                 self.clist.remove(username)
 
@@ -1501,11 +1486,7 @@ class ChatRoom:
     def count_users(self):
 
         numusers = len(self.users)
-        if numusers > 0:
-            self.LabelPeople.show()
-            self.LabelPeople.set_text(_("User List (%i)") % numusers)
-        else:
-            self.LabelPeople.hide()
+        self.LabelPeople.set_markup("<b>%d</b>" % numusers)
 
         if self.room in self.roomsctrl.rooms:
             iterator = self.roomsctrl.roomsmodel.get_iter_first()
@@ -1535,8 +1516,6 @@ class ChatRoom:
                 cellrenderer.set_property("weight", Pango.Weight.NORMAL)
                 cellrenderer.set_property("underline", Pango.Underline.NONE)
 
-        self.frame.cell_data_func(column, cellrenderer, model, iterator)
-
     def get_user_stats(self, user, avgspeed, files):
 
         if user not in self.users:
@@ -1563,7 +1542,7 @@ class ChatRoom:
 
         if user in self.tag_users:
             color = self.get_user_status_color(status)
-            self.changecolour(self.tag_users[user], color)
+            self.update_tag_visuals(self.tag_users[user], color)
 
         self.usersmodel.set(self.users[user], 0, img, 5, status)
 
@@ -1574,15 +1553,12 @@ class ChatRoom:
 
         self.usersmodel.set(self.users[user], 1, self.frame.get_flag_image(flag), 8, flag)
 
-    def makecolour(self, buffer, colour, username=None):
+    def create_tag(self, buffer, color, username=None):
 
-        colour = self.frame.np.config.sections["ui"][colour]
-        font = self.frame.np.config.sections["ui"]["chatfont"]
+        tag = buffer.create_tag()
 
-        tag = buffer.create_tag(font=font)
-
-        if colour:
-            tag.set_property("foreground", colour)
+        set_widget_color(tag, self.frame.np.config.sections["ui"][color])
+        set_widget_font(tag, self.frame.np.config.sections["ui"]["chatfont"])
 
         if username is not None:
 
@@ -1621,7 +1597,7 @@ class ChatRoom:
             self.popup_menu.set_user(user)
             me = (self.popup_menu.user is None or self.popup_menu.user == self.frame.np.config.sections["server"]["login"])
 
-            self.menu_add_to_list.set_active(user in [i[0] for i in self.frame.np.config.sections["server"]["userlist"]])
+            self.menu_add_to_list.set_active(user in (i[0] for i in self.frame.np.config.sections["server"]["userlist"]))
             self.menu_ban_user.set_active(user in self.frame.np.config.sections["server"]["banlist"])
             self.menu_ignore_user.set_active(user in self.frame.np.config.sections["server"]["ignorelist"])
             self.menu_block_user.set_active(self.frame.user_ip_is_blocked(user))
@@ -1635,26 +1611,28 @@ class ChatRoom:
 
         return True
 
-    def update_colours(self):
+    def update_visuals(self):
 
-        self.frame.change_list_font(self.UserList, self.frame.np.config.sections["ui"]["listfont"])
+        for widget in self.__dict__.values():
+            update_widget_visuals(widget, update_text_tags=False)
+
+        self.room_wall.update_visuals()
+
+    def create_tags(self):
 
         buffer = self.ChatScroll.get_buffer()
 
-        self.tag_remote = self.makecolour(buffer, "chatremote")
-        self.tag_local = self.makecolour(buffer, "chatlocal")
-        self.tag_me = self.makecolour(buffer, "chatme")
-        self.tag_hilite = self.makecolour(buffer, "chathilite")
+        self.tag_remote = self.create_tag(buffer, "chatremote")
+        self.tag_local = self.create_tag(buffer, "chatlocal")
+        self.tag_me = self.create_tag(buffer, "chatme")
+        self.tag_hilite = self.create_tag(buffer, "chathilite")
 
         self.tag_users = {}
         for user in self.tag_users:
             self.get_user_tag(user)
 
         logbuffer = self.RoomLog.get_buffer()
-        self.tag_log = self.makecolour(logbuffer, "chatremote")
-
-        self.frame.set_text_bg(self.ChatEntry)
-        self.frame.set_text_bg(self.room_wall.RoomWallEntry)
+        self.tag_log = self.create_tag(logbuffer, "chatremote")
 
     def get_user_status_color(self, status):
 
@@ -1670,20 +1648,13 @@ class ChatRoom:
 
         return color
 
-    def changecolour(self, tag, colour):
+    def update_tag_visuals(self, tag, color):
 
-        color = self.frame.np.config.sections["ui"][colour]
-
-        if color == "":
-            color = None
-
-        tag.set_property("foreground", color)
-
-        font = self.frame.np.config.sections["ui"]["chatfont"]
-        tag.set_property("font", font)
+        set_widget_color(tag, self.frame.np.config.sections["ui"][color])
+        set_widget_font(tag, self.frame.np.config.sections["ui"]["chatfont"])
 
         # Hotspots
-        if colour in ["useraway", "useronline", "useroffline"]:
+        if color in ("useraway", "useronline", "useroffline"):
 
             usernamestyle = self.frame.np.config.sections["ui"]["usernamestyle"]
 
@@ -1702,21 +1673,17 @@ class ChatRoom:
             else:
                 tag.set_property("underline", Pango.Underline.NONE)
 
-    def change_colours(self):
+    def update_tags(self):
 
-        self.changecolour(self.tag_remote, "chatremote")
-        self.changecolour(self.tag_local, "chatlocal")
-        self.changecolour(self.tag_me, "chatme")
-        self.changecolour(self.tag_hilite, "chathilite")
-        self.changecolour(self.tag_log, "chatremote")
+        self.update_tag_visuals(self.tag_remote, "chatremote")
+        self.update_tag_visuals(self.tag_local, "chatlocal")
+        self.update_tag_visuals(self.tag_me, "chatme")
+        self.update_tag_visuals(self.tag_hilite, "chathilite")
+        self.update_tag_visuals(self.tag_log, "chatremote")
 
         for user in self.tag_users:
             self.get_user_tag(user)
 
-        self.frame.set_text_bg(self.ChatEntry)
-        self.frame.set_text_bg(self.room_wall.RoomWallEntry)
-        self.frame.change_list_font(self.UserList, self.frame.np.config.sections["ui"]["listfont"])
-
     def on_leave(self, widget=None):
 
         if self.leaving:
@@ -1770,7 +1737,7 @@ class ChatRoom:
             del config["columns"]["chatrooms_widths"][self.room]
 
         for tag in self.tag_users.values():
-            self.changecolour(tag, "useroffline")
+            self.update_tag_visuals(tag, "useroffline")
 
         self.tickers.set_ticker([])
 
@@ -2037,7 +2004,7 @@ class ChatRooms(IconNotebook):
             self.frame.images,
             angle=config["ui"]["labelrooms"],
             tabclosers=config["ui"]["tabclosers"],
-            show_image=config["notifications"]["notification_tab_icons"],
+            show_hilite_image=config["notifications"]["notification_tab_icons"],
             reorderable=config["ui"]["tab_reorderable"],
             notebookraw=self.frame.ChatNotebookRaw
         )
@@ -2067,7 +2034,7 @@ class ChatRooms(IconNotebook):
 
             n = self.page_num(child)
             page = self.get_nth_page(n)
-            room = [room for room, tab in self.roomsctrl.joinedrooms.items() if tab.Main is page][0]
+            room = next(room for room, tab in self.roomsctrl.joinedrooms.items() if tab.Main is page)
 
             if event.button == 2:
                 self.roomsctrl.joinedrooms[room].on_leave(widget)
