@@ -27,6 +27,7 @@ from time import altzone
 from time import daylight
 
 from gi.repository import Gdk
+from gi.repository import Gio
 from gi.repository import GLib
 from gi.repository import GObject
 from gi.repository import Gtk
@@ -34,14 +35,17 @@ from gi.repository import Pango
 
 from pynicotine import slskmessages
 from pynicotine.gtkgui.chatrooms import get_completion
-from pynicotine.gtkgui.nowplaying import NowPlaying
 from pynicotine.gtkgui.utils import append_line
+from pynicotine.gtkgui.utils import expand_alias
 from pynicotine.gtkgui.utils import IconNotebook
+from pynicotine.gtkgui.utils import is_alias
+from pynicotine.gtkgui.utils import load_ui_elements
 from pynicotine.gtkgui.utils import PopupMenu
 from pynicotine.gtkgui.utils import scroll_bottom
 from pynicotine.gtkgui.utils import TextSearchBar
-from pynicotine.gtkgui.utils import expand_alias
-from pynicotine.gtkgui.utils import is_alias
+from pynicotine.gtkgui.utils import set_widget_color
+from pynicotine.gtkgui.utils import set_widget_font
+from pynicotine.gtkgui.utils import update_widget_visuals
 from pynicotine.logfacility import log
 from pynicotine.utils import clean_file
 from pynicotine.utils import version
@@ -72,7 +76,7 @@ class PrivateChats(IconNotebook):
             self.frame.images,
             angle=config["ui"]["labelprivate"],
             tabclosers=config["ui"]["tabclosers"],
-            show_image=config["notifications"]["notification_tab_icons"],
+            show_hilite_image=config["notifications"]["notification_tab_icons"],
             reorderable=config["ui"]["tab_reorderable"],
             show_status_image=config["ui"]["tab_status_icons"],
             notebookraw=self.frame.PrivatechatNotebookRaw
@@ -86,9 +90,9 @@ class PrivateChats(IconNotebook):
 
         self.notebook.connect("switch-page", self.on_switch_page)
 
-    def on_switch_page(self, notebook, page, page_num, force=0):
+    def on_switch_page(self, notebook, page, page_num, forceupdate=False):
 
-        if self.frame.MainNotebook.get_current_page() != self.frame.MainNotebook.page_num(self.frame.privatevbox) and not force:
+        if self.frame.MainNotebook.get_current_page() != self.frame.MainNotebook.page_num(self.frame.privatevbox) and not forceupdate:
             return
 
         for user, tab in list(self.users.items()):
@@ -136,15 +140,21 @@ class PrivateChats(IconNotebook):
             self.set_status_image(tab.Main, msg.status)
             tab.get_user_status(msg.status)
 
-    def send_message(self, user, text=None, direction=None, bytestring=False):
+    def send_message(self, user, text=None, show_user=False, bytestring=False):
 
         if user not in self.users:
             tab = PrivateChat(self, user)
             self.users[user] = tab
-            self.append_page(tab.Main, user, tab.on_close)
+
+            if not self.frame.np.config.sections["ui"]["tab_status_icons"]:
+                userlabel = "%s (%s)" % (user[:15], _("Offline"))
+            else:
+                userlabel = user
+
+            self.append_page(tab.Main, userlabel, tab.on_close)
             self.frame.np.queue.put(slskmessages.AddUser(user))
 
-        if direction:
+        if show_user:
             if self.get_current_page() != self.page_num(self.users[user].Main):
                 self.set_current_page(self.page_num(self.users[user].Main))
 
@@ -178,7 +188,7 @@ class PrivateChats(IconNotebook):
         me = (popup.user is None or popup.user == self.frame.np.config.sections["server"]["login"])
         items = popup.get_children()
 
-        items[6].set_active(user in [i[0] for i in self.frame.np.config.sections["server"]["userlist"]])
+        items[6].set_active(user in (i[0] for i in self.frame.np.config.sections["server"]["userlist"]))
         items[7].set_active(user in self.frame.np.config.sections["server"]["banlist"])
         items[8].set_active(user in self.frame.np.config.sections["server"]["ignorelist"])
         items[9].set_active(self.frame.user_ip_is_blocked(user))
@@ -194,7 +204,7 @@ class PrivateChats(IconNotebook):
 
             n = self.page_num(child)
             page = self.get_nth_page(n)
-            username = [user for user, tab in list(self.users.items()) if tab.Main is page][0]
+            username = next(user for user, tab in list(self.users.items()) if tab.Main is page)
 
             if event.button == 2:
                 self.users[username].on_close(widget)
@@ -218,6 +228,7 @@ class PrivateChats(IconNotebook):
             ip, port = self.frame.np.users[msg.user].addr
             if self.frame.np.ip_ignored(ip):
                 return
+
         elif newmessage:
             self.frame.np.queue.put(slskmessages.GetPeerAddress(msg.user))
             self.frame.np.private_message_queue_add(msg, text)
@@ -234,7 +245,7 @@ class PrivateChats(IconNotebook):
         self.request_changed(chat.Main)
 
         # Hilight main private chats Label
-        self.frame.request_icon(self.frame.PrivateChatTabLabel, chat.Main)
+        self.frame.request_tab_icon(self.frame.PrivateChatTabLabel)
 
         # Show notifications if the private chats notebook isn't selected,
         # the tab is not selected, or the main window isn't mapped
@@ -246,7 +257,8 @@ class PrivateChats(IconNotebook):
             if self.frame.np.config.sections["notifications"]["notification_popup_private_message"]:
                 self.frame.notifications.new_notification(
                     text,
-                    title=_("Private message from %s") % msg.user
+                    title=_("Private message from %s") % msg.user,
+                    priority=Gio.NotificationPriority.HIGH
                 )
 
         # SEND CLIENT VERSION to user if the following string is sent
@@ -262,9 +274,10 @@ class PrivateChats(IconNotebook):
 
         self.frame.np.pluginhandler.incoming_private_chat_notification(msg.user, text)
 
-    def update_colours(self):
+    def update_visuals(self):
         for chat in self.users.values():
-            chat.change_colours()
+            chat.update_visuals()
+            chat.update_tags()
 
     def remove_tab(self, tab):
 
@@ -292,15 +305,21 @@ class PrivateChats(IconNotebook):
             self.frame.np.config.sections["privatechat"]["users"].sort()
             for user in self.frame.np.config.sections["privatechat"]["users"]:
                 if user not in self.users:
-                    self.send_message(user, None, 1)
+                    self.send_message(user, show_user=True)
 
     def conn_close(self):
 
         self.connected = 0
+
         for user in self.users:
             self.users[user].conn_close()
             tab = self.users[user]
-            self.set_text(tab.Main, "%s (%s)" % (user[:15], _("Offline")))
+
+            if not self.frame.np.config.sections["ui"]["tab_status_icons"]:
+                self.set_text(tab.Main, "%s (%s)" % (user[:15], _("Offline")))
+            else:
+                self.set_text(tab.Main, user)
+
             self.set_status_image(tab.Main, 0)
             tab.get_user_status(0)
 
@@ -335,23 +354,7 @@ class PrivateChat:
         # We should reference the user as soon as possible
         self.chats.users[self.user] = self
 
-        builder = Gtk.Builder()
-
-        builder.set_translation_domain('nicotine')
-        builder.add_from_file(os.path.join(os.path.dirname(os.path.realpath(__file__)), "ui", "privatechat.ui"))
-
-        self.private_chat_tab = builder.get_object("PrivateChatTab")
-
-        for i in builder.get_objects():
-            try:
-                self.__dict__[Gtk.Buildable.get_name(i)] = i
-            except TypeError:
-                pass
-
-        self.private_chat_tab.remove(self.Main)
-        self.private_chat_tab.destroy()
-
-        builder.connect_signals(self)
+        load_ui_elements(self, os.path.join(self.frame.gui_dir, "ui", "privatechat.ui"))
 
         self.autoreplied = 0
         self.offlinemessage = 0
@@ -362,6 +365,9 @@ class PrivateChat:
         TextSearchBar(self.ChatScroll, self.SearchBar, self.SearchEntry)
 
         # Spell Check
+        if self.frame.spell_checker is None:
+            self.frame.init_spell_checker()
+
         if self.frame.spell_checker and self.frame.np.config.sections["ui"]["spellcheck"]:
             from gi.repository import Gspell
             spell_buffer = Gspell.EntryBuffer.get_from_gtk_entry_buffer(self.ChatLine.get_buffer())
@@ -416,7 +422,8 @@ class PrivateChat:
         self.ChatScroll.connect("button_press_event", self.on_popup_menu)
         self.ChatScroll.connect("key_press_event", self.on_popup_menu)
 
-        self.update_colours()
+        self.create_tags()
+        self.update_visuals()
 
         self.chats.update_completions()
 
@@ -448,14 +455,14 @@ class PrivateChat:
     def login(self):
         timestamp_format = self.frame.np.config.sections["logging"]["private_timestamp"]
         append_line(self.ChatScroll, _("--- reconnected ---"), self.tag_hilite, timestamp_format=timestamp_format)
-        self.change_colours()
+        self.update_tags()
 
     def conn_close(self):
         timestamp_format = self.frame.np.config.sections["logging"]["private_timestamp"]
         append_line(self.ChatScroll, _("--- disconnected ---"), self.tag_hilite, timestamp_format=timestamp_format)
         self.status = -1
         self.offlinemessage = 0
-        self.change_colours()
+        self.update_tags()
 
     def on_popup_menu(self, widget, event):
 
@@ -480,7 +487,7 @@ class PrivateChat:
         items = self.popup_menu_user.get_children()
         me = (self.popup_menu_user.user is None or self.popup_menu_user.user == self.frame.np.config.sections["server"]["login"])
 
-        items[6].set_active(self.user in [i[0] for i in self.frame.np.config.sections["server"]["userlist"]])
+        items[6].set_active(self.user in (i[0] for i in self.frame.np.config.sections["server"]["userlist"]))
         items[7].set_active(self.user in self.frame.np.config.sections["server"]["banlist"])
         items[8].set_active(self.user in self.frame.np.config.sections["server"]["ignorelist"])
         items[9].set_active(self.frame.user_ip_is_blocked(self.user))
@@ -514,7 +521,7 @@ class PrivateChat:
 
     def show_message(self, text, newmessage=True, timestamp=None):
 
-        self.update_colours()
+        self.create_tags()
 
         if text[:4] == "/me ":
             line = "* %s %s" % (self.user, self.frame.censor_chat(text[4:]))
@@ -605,7 +612,7 @@ class PrivateChat:
 
         if self.PeerPrivateMessages.get_active():
             # not in the soulseek protocol
-            self.frame.np.process_request_to_peer(self.user, slskmessages.PMessageUser(None, my_username, payload))
+            self.frame.np.send_message_to_peer(self.user, slskmessages.PMessageUser(None, my_username, payload))
         else:
             self.frame.np.queue.put(slskmessages.MessageUser(self.user, payload))
 
@@ -673,13 +680,12 @@ class PrivateChat:
         elif cmd == "/ip":
             if args:
                 user = args
-                if user not in self.frame.np.ip_requested:
-                    self.frame.np.ip_requested.append(user)
+                self.frame.np.ip_requested.add(user)
                 self.frame.np.queue.put(slskmessages.GetPeerAddress(user))
 
         elif cmd == "/pm":
             if realargs:
-                self.frame.privatechats.send_message(realargs, None, 1)
+                self.frame.privatechats.send_message(realargs, show_user=True)
 
         elif cmd in ["/m", "/msg"]:
             if realargs:
@@ -742,7 +748,7 @@ class PrivateChat:
 
         elif cmd == "/ctcpversion":
             if args:
-                self.frame.privatechats.send_message(args, CTCP_VERSION, 1, bytestring=True)
+                self.frame.privatechats.send_message(args, CTCP_VERSION, show_user=True, bytestring=True)
 
         elif cmd in ["/clear", "/cl"]:
             self.ChatScroll.get_buffer().set_text("")
@@ -751,7 +757,7 @@ class PrivateChat:
             self.frame.on_away(None)
 
         elif cmd in ["/q", "/quit", "/exit"]:
-            self.frame.on_exit(None)
+            self.frame.on_quit(None)
             return
 
         elif cmd in ["/c", "/close"]:
@@ -784,30 +790,28 @@ class PrivateChat:
         widget.set_text("")
 
     def display_now_playing(self):
-        if self.frame.now_playing is None:
-            self.frame.now_playing = NowPlaying(self.frame)
+        self.frame.now_playing.display_now_playing(callback=self.send_message)
 
-        self.frame.now_playing.display_now_playing(None, 0, self.send_message)
+    def update_visuals(self):
 
-    def makecolour(self, buffer, colour):
+        for widget in self.__dict__.values():
+            update_widget_visuals(widget, update_text_tags=False)
 
-        colour = self.frame.np.config.sections["ui"][colour]
-        font = self.frame.np.config.sections["ui"]["chatfont"]
+    def create_tag(self, buffer, color):
 
-        tag = buffer.create_tag(font=font)
-
-        if colour:
-            tag.set_property("foreground", colour)
+        tag = buffer.create_tag()
+        set_widget_color(tag, self.frame.np.config.sections["ui"][color])
+        set_widget_font(tag, self.frame.np.config.sections["ui"]["chatfont"])
 
         return tag
 
-    def update_colours(self):
+    def create_tags(self):
 
         buffer = self.ChatScroll.get_buffer()
-        self.tag_remote = self.makecolour(buffer, "chatremote")
-        self.tag_local = self.makecolour(buffer, "chatlocal")
-        self.tag_me = self.makecolour(buffer, "chatme")
-        self.tag_hilite = self.makecolour(buffer, "chathilite")
+        self.tag_remote = self.create_tag(buffer, "chatremote")
+        self.tag_local = self.create_tag(buffer, "chatlocal")
+        self.tag_me = self.create_tag(buffer, "chatme")
+        self.tag_hilite = self.create_tag(buffer, "chathilite")
 
         if self.status == 1 and self.frame.np.config.sections["ui"]["showaway"]:
             statuscolor = "useraway"
@@ -816,15 +820,15 @@ class PrivateChat:
         else:
             statuscolor = "useroffline"
 
-        self.tag_username = self.makecolour(buffer, statuscolor)
+        self.tag_username = self.create_tag(buffer, statuscolor)
 
         if self.chats.connected:
             if self.frame.away and self.frame.np.config.sections["ui"]["showaway"]:
-                self.tag_my_username = self.makecolour(buffer, "useraway")
+                self.tag_my_username = self.create_tag(buffer, "useraway")
             else:
-                self.tag_my_username = self.makecolour(buffer, "useronline")
+                self.tag_my_username = self.create_tag(buffer, "useronline")
         else:
-            self.tag_my_username = self.makecolour(buffer, "useroffline")
+            self.tag_my_username = self.create_tag(buffer, "useroffline")
 
         usernamestyle = self.frame.np.config.sections["ui"]["usernamestyle"]
 
@@ -849,22 +853,12 @@ class PrivateChat:
             self.tag_username.set_property("underline", Pango.Underline.NONE)
             self.tag_my_username.set_property("underline", Pango.Underline.NONE)
 
-        self.frame.set_text_bg(self.ChatLine)
-        self.frame.set_text_bg(self.PeerPrivateMessages)
+    def update_tag_visuals(self, tag, color):
 
-    def changecolour(self, tag, colour):
+        set_widget_color(tag, self.frame.np.config.sections["ui"][color])
+        set_widget_font(tag, self.frame.np.config.sections["ui"]["chatfont"])
 
-        color = self.frame.np.config.sections["ui"][colour]
-
-        if color == "":
-            color = None
-
-        tag.set_property("foreground", color)
-
-        font = self.frame.np.config.sections["ui"]["chatfont"]
-        tag.set_property("font", font)
-
-        if colour in ["useraway", "useronline", "useroffline"]:
+        if color in ("useraway", "useronline", "useroffline"):
 
             usernamestyle = self.frame.np.config.sections["ui"]["usernamestyle"]
 
@@ -883,25 +877,23 @@ class PrivateChat:
             else:
                 tag.set_property("underline", Pango.Underline.NONE)
 
-    def change_colours(self):
+    def update_tags(self):
 
-        self.changecolour(self.tag_remote, "chatremote")
-        self.changecolour(self.tag_local, "chatlocal")
-        self.changecolour(self.tag_me, "chatme")
-        self.changecolour(self.tag_hilite, "chathilite")
+        self.update_tag_visuals(self.tag_remote, "chatremote")
+        self.update_tag_visuals(self.tag_local, "chatlocal")
+        self.update_tag_visuals(self.tag_me, "chatme")
+        self.update_tag_visuals(self.tag_hilite, "chathilite")
 
         color = self.get_user_status_color(self.status)
-        self.changecolour(self.tag_username, color)
+        self.update_tag_visuals(self.tag_username, color)
 
         if self.chats.connected:
             if self.frame.away and self.frame.np.config.sections["ui"]["showaway"]:
-                self.changecolour(self.tag_my_username, "useraway")
+                self.update_tag_visuals(self.tag_my_username, "useraway")
             else:
-                self.changecolour(self.tag_my_username, "useronline")
+                self.update_tag_visuals(self.tag_my_username, "useronline")
         else:
-            self.changecolour(self.tag_my_username, "useroffline")
-
-        self.frame.set_text_bg(self.ChatLine)
+            self.update_tag_visuals(self.tag_my_username, "useroffline")
 
     def get_user_status_color(self, status):
 
@@ -924,7 +916,7 @@ class PrivateChat:
 
         self.status = status
         color = self.get_user_status_color(self.status)
-        self.changecolour(self.tag_username, color)
+        self.update_tag_visuals(self.tag_username, color)
 
     def on_close(self, widget):
 
