@@ -25,15 +25,12 @@
 import re
 import sys
 import time
-import types
-import urllib.parse
-import webbrowser
 
 from gettext import gettext as _
 
 from gi.repository import Gdk
+from gi.repository import Gio
 from gi.repository import GLib
-from gi.repository import GObject
 from gi.repository import Gtk
 from gi.repository import Pango
 
@@ -44,13 +41,7 @@ from pynicotine.logfacility import log
 from pynicotine.utils import execute_command
 
 
-DECIMALSEP = ""
-
 URL_RE = re.compile("(\\w+\\://[^\\s]+)|(www\\.\\w+\\.\\w+.*?)|(mailto\\:[^\\s]+)")
-PROTOCOL_HANDLERS = {}
-CATCH_URLS = 0
-HUMANIZE_URLS = 0
-USERNAMEHOTSPOTS = 0
 NICOTINE = None
 
 
@@ -113,6 +104,24 @@ def show_country_tooltip(widget, x, y, tooltip, sourcecolumn, stripprefix='flag_
     return True
 
 
+def load_ui_elements(ui_class, filename):
+    builder = Gtk.Builder()
+
+    builder.set_translation_domain('nicotine')
+    builder.add_from_file(filename)
+
+    for i in builder.get_objects():
+        try:
+            obj_name = Gtk.Buildable.get_name(i)
+
+            if not obj_name.startswith("_"):
+                ui_class.__dict__[obj_name] = i
+        except TypeError:
+            pass
+
+    builder.connect_signals(ui_class)
+
+
 def fill_file_grouping_combobox(combobox):
     grouplist = Gtk.ListStore(str)
     groups = [
@@ -171,35 +180,42 @@ def initialise_columns(treeview, *args):
             renderer.set_padding(10, 3)
 
             column = Gtk.TreeViewColumn(c[0], renderer, text=i)
+
         elif c[2] == "center":
             renderer = Gtk.CellRendererText()
             renderer.set_property("xalign", 0.5)
 
             column = Gtk.TreeViewColumn(c[0], renderer, text=i)
+
         elif c[2] == "number":
             renderer = Gtk.CellRendererText()
             renderer.set_property("xalign", 0.9)
 
             column = Gtk.TreeViewColumn(c[0], renderer, text=i)
             column.set_alignment(0.9)
+
         elif c[2] == "edit":
             renderer = Gtk.CellRendererText()
             renderer.set_padding(10, 3)
             renderer.set_property('editable', True)
             column = Gtk.TreeViewColumn(c[0], renderer, text=i)
+
         elif c[2] == "combo":
             renderer = Gtk.CellRendererCombo()
             renderer.set_padding(10, 3)
             renderer.set_property('text-column', 0)
             renderer.set_property('editable', True)
             column = Gtk.TreeViewColumn(c[0], renderer, text=i)
+
         elif c[2] == "progress":
             renderer = Gtk.CellRendererProgress()
             column = Gtk.TreeViewColumn(c[0], renderer, value=i)
+
         elif c[2] == "toggle":
             renderer = Gtk.CellRendererToggle()
             column = Gtk.TreeViewColumn(c[0], renderer, active=i)
             renderer.set_property("xalign", 0.5)
+
         else:
             renderer = Gtk.CellRendererPixbuf()
             column = Gtk.TreeViewColumn(c[0], renderer, pixbuf=i)
@@ -207,6 +223,7 @@ def initialise_columns(treeview, *args):
         if c[1] == -1:
             column.set_resizable(False)
             column.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
+
         else:
             column.set_resizable(True)
             if c[1] == 0:
@@ -223,14 +240,15 @@ def initialise_columns(treeview, *args):
             foreground = c[4][0]
             background = c[4][1]
 
-            if foreground == "":
-                foreground = None
+            if foreground:
+                renderer.set_property("foreground", foreground)
+            else:
+                renderer.set_property("foreground-set", False)
 
-            if background == "":
-                background = None
-
-            renderer.set_property("foreground", foreground)
-            renderer.set_property("background", background)
+            if background:
+                renderer.set_property("background", background)
+            else:
+                renderer.set_property("background-set", False)
 
         column.set_reorderable(False)
         column.set_widget(Gtk.Label.new(c[0]))
@@ -331,7 +349,7 @@ def header_toggle(menuitem, columns, index):
     prev_column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
     prev_column.set_resizable(True)
 
-    NICOTINE.SaveColumns()
+    NICOTINE.save_columns()
 
 
 def set_treeview_selected_row(treeview, event):
@@ -352,6 +370,20 @@ def set_treeview_selected_row(treeview, event):
         selection.unselect_all()
 
 
+def open_file_path(file_path, command=None):
+    """ Currently used to either open a folder or play an audio file
+    Tries to run a user-specified command first, and falls back to
+    the system default. """
+
+    if command and "$" in command:
+        execute_command(command, file_path)
+    else:
+        try:
+            Gio.AppInfo.launch_default_for_uri("file:///" + file_path)
+        except GLib.Error as error:
+            log.add_warning(_("Failed to open folder: %s"), str(error))
+
+
 def scroll_bottom(widget):
     va = widget.get_vadjustment()
     try:
@@ -377,19 +409,21 @@ def open_uri(uri, window):
 
     # Situation 1, user defined a way of handling the protocol
     protocol = uri[:uri.find(":")]
-    if protocol in PROTOCOL_HANDLERS:
-        if isinstance(PROTOCOL_HANDLERS[protocol], types.MethodType):
-            PROTOCOL_HANDLERS[protocol](uri.strip())
+    protocol_handlers = NICOTINE.np.config.sections["urls"]["protocols"]
+
+    if protocol in protocol_handlers and protocol_handlers[protocol]:
+        try:
+            execute_command(protocol_handlers[protocol], uri)
             return
-        if PROTOCOL_HANDLERS[protocol]:
-            try:
-                execute_command(PROTOCOL_HANDLERS[protocol], uri)
-                return
-            except RuntimeError as e:
-                log.add_warning("%s", e)
+        except RuntimeError as e:
+            log.add_warning("%s", e)
+
+    if protocol == "slsk":
+        on_soul_seek_uri(uri.strip())
 
     # Situation 2, user did not define a way of handling the protocol
-    if sys.platform == "win32" and webbrowser:
+    if sys.platform == "win32":
+        import webbrowser
         webbrowser.open(uri)
         return
 
@@ -400,6 +434,21 @@ def open_uri(uri, window):
         Gtk.show_uri(screen, uri, Gdk.CURRENT_TIME)
 
 
+def on_soul_seek_uri(url):
+    import urllib.parse
+
+    try:
+        user, file = urllib.parse.unquote(url[7:]).split("/", 1)
+
+        if file[-1] == "/":
+            NICOTINE.np.send_message_to_peer(user, slskmessages.FolderContentsRequest(None, file[:-1].replace("/", "\\")))
+        else:
+            NICOTINE.np.transfers.get_file(user, file.replace("/", "\\"), "")
+
+    except Exception:
+        log.add(_("Invalid SoulSeek meta-url: %s"), url)
+
+
 def append_line(textview, line, tag=None, timestamp=None, showstamp=True, timestamp_format="%H:%M:%S", username=None, usertag=None, scroll=True):
 
     if type(line) not in (type(""), type("")):
@@ -429,7 +478,7 @@ def append_line(textview, line, tag=None, timestamp=None, showstamp=True, timest
 
     def _usertag(buffer, section):
         # Tag usernames with popup menu creating tag, and away/online/offline colors
-        if USERNAMEHOTSPOTS and username is not None and usertag is not None:
+        if NICOTINE.np.config.sections["ui"]["usernamehotspots"] and username is not None and usertag is not None:
             np = re.compile(re.escape(str(username)))
             match = np.search(section)
             if match is not None:
@@ -484,14 +533,15 @@ def append_line(textview, line, tag=None, timestamp=None, showstamp=True, timest
     # Match first url
     match = URL_RE.search(line)
     # Highlight urls, if found and tag them
-    while CATCH_URLS and match:
+    while NICOTINE.np.config.sections["urls"]["urlcatching"] and match:
         start = line[:match.start()]
         _usertag(buffer, start)
         url = match.group()
         urltag = _makeurltag(buffer, url)
         line = line[match.end():]
 
-        if url.startswith("slsk://") and HUMANIZE_URLS:
+        if url.startswith("slsk://") and NICOTINE.np.config.sections["urls"]["humanizeurls"]:
+            import urllib.parse
             url = urllib.parse.unquote(url)
 
         _append(buffer, url, urltag)
@@ -517,13 +567,10 @@ class BuddiesComboBox:
 
         self.combobox = combo_box
 
-        self.store = Gtk.ListStore(GObject.TYPE_STRING)
+        self.store = Gtk.ListStore(str)
         self.combobox.set_model(self.store)
         self.combobox.set_entry_text_column(0)
 
-        self.store.set_default_sort_func(lambda *args: -1)
-        self.store.set_sort_column_id(-1, Gtk.SortType.ASCENDING)
-
         self.combobox.show()
 
     def fill(self):
@@ -536,8 +583,6 @@ class BuddiesComboBox:
         for user in self.frame.np.config.sections["server"]["userlist"]:
             self.items[user[0]] = self.store.append([user[0]])
 
-        self.store.set_sort_column_id(0, Gtk.SortType.ASCENDING)
-
     def append(self, item):
 
         if item in self.items:
@@ -554,39 +599,34 @@ class BuddiesComboBox:
 
 class ImageLabel(Gtk.Box):
 
-    def __init__(self, label="", image=None, onclose=None, closebutton=False, angle=0, show_image=True, statusimage=None, show_status_image=False):
+    def __init__(self, label="", onclose=None, closebutton=False, angle=0, hilite_image=None, show_hilite_image=True, status_image=None, show_status_image=False):
 
         Gtk.Box.__init__(self)
 
         self.closebutton = closebutton
         self.angle = angle
-        self._show_image = show_image
-        self._show_status_image = show_status_image
-        self.notify = 0
-
-        self._entered = 0
-        self._pressed = 0
 
         self.onclose = onclose
-        self.status_img = None
-        self.statusimage = Gtk.Image()
+
         self.label = Gtk.Label()
-        self.text = label
-
-        self.set_text(self.text)
-
         self.label.set_angle(angle)
         self.label.show()
 
-        if self._show_status_image:
-            self.set_status_image(statusimage)
-            self.statusimage.show()
+        self.text = label
+        self.set_text(self.text)
 
-        self.image = Gtk.Image()
-        self.set_image(image)
+        self.status_image = Gtk.Image()
+        self.status_pixbuf = None
 
-        if self._show_image:
-            self.image.show()
+        if show_status_image:
+            self.set_status_image(status_image)
+            self.status_image.show()
+
+        self.hilite_image = Gtk.Image()
+        self.hilite_pixbuf = None
+
+        if show_hilite_image:
+            self.set_hilite_image(hilite_image)
 
         self._pack_children()
         self._order_children()
@@ -612,10 +652,12 @@ class ImageLabel(Gtk.Box):
         self.box.set_spacing(2)
         self.add(self.box)
         self.box.show()
+        self.status_image.set_margin_end(5)
+        self.hilite_image.set_margin_start(5)
 
-        self.box.pack_start(self.statusimage, False, False, 0)
+        self.box.pack_start(self.status_image, False, False, 0)
         self.box.pack_start(self.label, True, True, 0)
-        self.box.pack_start(self.image, False, False, 0)
+        self.box.pack_start(self.hilite_image, False, False, 0)
 
         if self.closebutton and self.onclose is not None:
             self._add_close_button()
@@ -625,17 +667,17 @@ class ImageLabel(Gtk.Box):
         if self.angle == 90:
             if "button" in self.__dict__ and self.closebutton != 0:
                 self.box.reorder_child(self.button, 0)
-                self.box.reorder_child(self.image, 1)
+                self.box.reorder_child(self.hilite_image, 1)
                 self.box.reorder_child(self.label, 2)
-                self.box.reorder_child(self.statusimage, 3)
+                self.box.reorder_child(self.status_image, 3)
             else:
-                self.box.reorder_child(self.image, 0)
+                self.box.reorder_child(self.hilite_image, 0)
                 self.box.reorder_child(self.label, 1)
-                self.box.reorder_child(self.statusimage, 2)
+                self.box.reorder_child(self.status_image, 2)
         else:
-            self.box.reorder_child(self.statusimage, 0)
+            self.box.reorder_child(self.status_image, 0)
             self.box.reorder_child(self.label, 1)
-            self.box.reorder_child(self.image, 2)
+            self.box.reorder_child(self.hilite_image, 2)
 
             if "button" in self.__dict__ and self.closebutton != 0:
                 self.box.reorder_child(self.button, 3)
@@ -650,13 +692,11 @@ class ImageLabel(Gtk.Box):
 
         self._order_children()
 
-    def show_image(self, show=True):
-        self._show_image = show
-
-        if self._show_image:
-            self.image.show()
+    def show_hilite_image(self, show=True):
+        if show and self.get_hilite_image() is not None:
+            self.hilite_image.show()
         else:
-            self.image.hide()
+            self.hilite_image.hide()
 
     def set_angle(self, angle):
         self.angle = angle
@@ -670,10 +710,11 @@ class ImageLabel(Gtk.Box):
         if "button" in self.__dict__:
             return
 
+        close_image = Gtk.Image()
+        close_image.set_from_icon_name("window-close-symbolic", Gtk.IconSize.MENU)
+
         self.button = Gtk.Button()
-        img = Gtk.Image()
-        img.set_from_icon_name("window-close-symbolic", Gtk.IconSize.MENU)
-        self.button.add(img)
+        self.button.add(close_image)
 
         if self.onclose is not None:
             self.button.connect("clicked", self.onclose)
@@ -692,10 +733,6 @@ class ImageLabel(Gtk.Box):
         del self.button
 
     def set_text_color(self, notify=None, text=None):
-        if notify is None:
-            notify = self.notify
-        else:
-            self.notify = notify
 
         color = NICOTINE.np.config.sections["ui"]["tab_default"]
 
@@ -719,28 +756,32 @@ class ImageLabel(Gtk.Box):
         else:
             self.label.set_markup("<span foreground=\"%s\">%s</span>" % (color, self.text.replace("<", "&lt;").replace(">", "&gt;").replace("&", "&amp;")))
 
-    def set_image(self, img):
-        self.img = img
-        self.image.set_from_pixbuf(img)
+    def set_hilite_image(self, pixbuf):
+        self.hilite_pixbuf = pixbuf
+        self.hilite_image.set_from_pixbuf(pixbuf)
 
-    def get_image(self):
-        return self.img
+        self.show_hilite_image()
 
-    def set_status_image(self, img):
-        if img is self.status_img:
+    def get_hilite_image(self):
+        return self.hilite_pixbuf
+
+    def set_status_image(self, pixbuf):
+        if pixbuf is self.status_pixbuf:
             return
 
-        if NICOTINE:
-            if NICOTINE.np.config.sections["ui"]["tab_status_icons"]:
-                self.statusimage.show()
-            else:
-                self.statusimage.hide()
+        if NICOTINE.np.config.sections["ui"]["tab_status_icons"]:
+            self.status_image.show()
+        else:
+            self.status_image.hide()
 
-        self.status_img = img
-        self.statusimage.set_from_pixbuf(img)
+        self.status_pixbuf = pixbuf
+        self.status_image.set_from_pixbuf(pixbuf)
 
     def get_status_image(self):
-        return self.status_img
+        return self.status_pixbuf
+
+    def set_icon(self, icon_name):
+        self.status_image.set_from_icon_name(icon_name, Gtk.IconSize.BUTTON)
 
     def set_text(self, lbl):
         self.set_text_color(notify=None, text=lbl)
@@ -756,52 +797,71 @@ class IconNotebook:
     - You can choose the label orientation (angle).
     """
 
-    def __init__(self, images, angle=0, tabclosers=False, show_image=True, reorderable=True, show_status_image=False, notebookraw=None):
+    def __init__(self, images, angle=0, tabclosers=False, show_hilite_image=True, reorderable=True, show_status_image=False, notebookraw=None):
 
         # We store the real Gtk.Notebook object
         self.notebook = notebookraw
-        self.notebook.set_show_border(True)
+
+        css_provider = Gtk.CssProvider()
+        css_provider.load_from_data(
+            b"""
+            .notebook {
+                border-left: none;
+                border-right: none;
+                border-bottom: none
+            }
+
+            """
+        )
+        context = self.notebook.get_style_context()
+        context.add_provider(css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
+        context.add_class("notebook")
 
         self.tabclosers = tabclosers
         self.reorderable = reorderable
 
         self.images = images
-        self._show_image = show_image
+        self._show_hilite_image = show_hilite_image
         self._show_status_image = show_status_image
 
-        self.pages = []
-
         self.notebook.connect("switch-page", self.dismiss_icon)
-        self.notebook.connect("key_press_event", self.on_key_press)
 
         self.angle = angle
 
+    def get_labels(self, page):
+        tab_label = self.notebook.get_tab_label(page).get_child()
+        menu_label = self.notebook.get_menu_label(page)
+
+        return tab_label, menu_label
+
     def set_reorderable(self, reorderable):
 
         self.reorderable = reorderable
 
-        for data in self.pages:
-            page, label_tab, status, label_tab_menu = data
-            try:
-                self.notebook.set_tab_reorderable(page, self.reorderable)
-            except Exception:
-                pass
+        for page in self.notebook.get_children():
+            self.notebook.set_tab_reorderable(page, self.reorderable)
 
     def set_tab_closers(self, closers):
 
         self.tabclosers = closers
 
-        for data in self.pages:
-            page, label_tab, status, label_tab_menu = data
-            label_tab.set_onclose(self.tabclosers)
+        for page in self.notebook.get_children():
+            tab_label, menu_label = self.get_labels(page)
 
-    def show_images(self, show_image=True):
+            tab_label.set_onclose(self.tabclosers)
 
-        self._show_image = show_image
+    def show_hilite_images(self, show_image=True):
 
-        for data in self.pages:
-            page, label_tab, status, label_tab_menu = data
-            label_tab.show_image(self._show_image)
+        self._show_hilite_image = show_image
+
+        for page in self.notebook.get_children():
+            tab_label, menu_label = self.get_labels(page)
+
+            tab_label.show_hilite_image(self._show_hilite_image)
+
+    def show_status_images(self, show_image=True):
+
+        self._show_status_image = show_image
 
     def set_tab_angle(self, angle):
 
@@ -810,37 +870,23 @@ class IconNotebook:
 
         self.angle = angle
 
-        for data in self.pages:
-            page, label_tab, status, label_tab_menu = data
-            label_tab.set_angle(angle)
+        for page in self.notebook.get_children():
+            tab_label, menu_label = self.get_labels(page)
+
+            tab_label.set_angle(angle)
 
     def set_tab_pos(self, pos):
         self.notebook.set_tab_pos(pos)
 
-    def on_key_press(self, widget, event):
-
-        if event.state & (Gdk.ModifierType.MOD1_MASK | Gdk.ModifierType.CONTROL_MASK) != Gdk.ModifierType.MOD1_MASK:
-            return False
-
-        if event.keyval in [Gdk.keyval_from_name("Up"), Gdk.keyval_from_name("Left")]:
-            self.prev_page()
-        elif event.keyval in [Gdk.keyval_from_name("Down"), Gdk.keyval_from_name("Right")]:
-            self.next_page()
-        else:
-            return False
-
-        widget.stop_emission_by_name("key_press_event")
-
-        return True
-
     def append_page(self, page, label, onclose=None, angle=0, fulltext=None):
 
         self.set_tab_angle(angle)
         closebutton = self.tabclosers
 
         label_tab = ImageLabel(
-            label, self.images["empty"], onclose, closebutton=closebutton,
-            angle=angle, show_image=self._show_image, statusimage=None,
+            label, onclose, closebutton=closebutton, angle=angle,
+            show_hilite_image=self._show_hilite_image,
+            status_image=self.images["offline"],
             show_status_image=self._show_status_image
         )
 
@@ -850,9 +896,7 @@ class IconNotebook:
         label_tab.set_tooltip_text(fulltext)
 
         # menu for all tabs
-        label_tab_menu = ImageLabel(label, self.images["empty"])
-
-        self.pages.append([page, label_tab, 0, label_tab_menu])
+        label_tab_menu = ImageLabel(label)
 
         eventbox = Gtk.EventBox()
         eventbox.set_visible_window(False)
@@ -871,76 +915,59 @@ class IconNotebook:
 
     def remove_page(self, page):
 
-        for i in self.pages[:]:
-            if i[0] == page:
-                Gtk.Notebook.remove_page(self.notebook, self.page_num(page))
-                i[1].destroy()
-                i[3].destroy()
-                self.pages.remove(i)
-
-                break
+        Gtk.Notebook.remove_page(self.notebook, self.page_num(page))
 
         if self.notebook.get_n_pages() == 0:
             self.notebook.set_show_tabs(False)
 
-    def on_focused(self, item):
-        self.frame.notifications.clear_page(self, item)
-
     def on_tab_click(self, widget, event, child):
         # Dummy implementation
         pass
 
     def set_status_image(self, page, status):
 
+        tab_label, menu_label = self.get_labels(page)
         image = self.images[("offline", "away", "online")[status]]
 
-        for i in self.pages:
-            if page == i[0]:
-                i[1].set_status_image(image)
-                i[3].set_status_image(image)
-                return
+        tab_label.set_status_image(image)
+        menu_label.set_status_image(image)
 
-    def set_image(self, page, status):
+    def set_hilite_image(self, page, status):
 
-        image = self.images[("empty", "hilite3", "hilite")[status]]
+        tab_label, menu_label = self.get_labels(page)
+        image = None
 
-        for i in self.pages:
-            if page == i[0]:
+        if status > 0:
+            image = self.images[("hilite3", "hilite")[status - 1]]
 
-                if status == 1 and i[2] == 2:
-                    return
+        if status == 1 and tab_label.get_hilite_image() == self.images["hilite"]:
+            # Chat mentions have priority over normal notifications
+            return
 
-                if i[2] != status:
-                    i[1].set_image(image)
-                    i[3].set_image(image)
-                    i[2] = status
-
-                return
+        tab_label.set_hilite_image(image)
+        menu_label.set_hilite_image(image)
 
     def set_text(self, page, label):
 
-        for i in self.pages:
-            if i[0] == page:
-                i[1].set_text(label)
-                i[3].set_text(label)
-                return
+        tab_label, menu_label = self.get_labels(page)
 
-    def set_text_colors(self, color=None):
+        tab_label.set_text(label)
+        menu_label.set_text(label)
 
-        for i in self.pages:
-            i[1].set_text_color(color)
+    def set_text_colors(self, status):
 
-    def set_text_color(self, page, color=None):
+        for page in self.notebook.get_children():
+            self.set_text_color(page, status)
 
-        for i in self.pages:
-            if i[0] == page:
-                i[1].set_text_color(color)
-                return
+    def set_text_color(self, page, status):
+
+        tab_label, menu_label = self.get_labels(page)
+        tab_label.set_text_color(status)
 
     def dismiss_icon(self, notebook, page, page_num):
 
-        self.set_image(page, 0)
-        self.set_text_color(page, 0)
+        self.set_hilite_image(page, status=0)
+        self.set_text_color(page, status=0)
 
     def request_hilite(self, page):
 
@@ -948,8 +975,8 @@ class IconNotebook:
         if current == page:
             return
 
-        self.set_image(page, 2)
-        self.set_text_color(page, 2)
+        self.set_hilite_image(page, status=2)
+        self.set_text_color(page, status=2)
 
     def request_changed(self, page):
 
@@ -957,8 +984,8 @@ class IconNotebook:
         if current == page:
             return
 
-        self.set_image(page, 1)
-        self.set_text_color(page, 1)
+        self.set_hilite_image(page, status=1)
+        self.set_text_color(page, status=1)
 
     def get_current_page(self):
         return self.notebook.get_current_page()
@@ -982,6 +1009,38 @@ class IconNotebook:
         self.notebook.show()
 
 
+class InfoBar:
+    """ Wrapper for setting up a GtkInfoBar """
+
+    def __init__(self, info_bar, message_type):
+
+        self.info_bar = info_bar
+        self.info_bar.set_message_type(message_type)
+        self.info_bar.set_show_close_button(True)
+        self.info_bar.connect("response", self._hide)
+
+        self.set_visible(False)
+
+    def _hide(self, *args):
+        self.set_visible(False)
+
+    def set_visible(self, visible):
+
+        try:
+            self.info_bar.set_revealed(visible)
+
+        except AttributeError:
+            # Older GTK version
+            self.info_bar.set_visible(visible)
+
+    def show_message(self, message):
+
+        label = self.info_bar.get_content_area().get_children()[0]
+        label.set_text(message)
+
+        self.set_visible(True)
+
+
 class PopupMenu(Gtk.Menu):
 
     def __init__(self, frame=None, shouldattach=True):
@@ -1085,17 +1144,15 @@ class PopupMenu(Gtk.Menu):
     def on_search_user(self, widget):
         self.frame.SearchMethod.set_active_iter(self.frame.searchmethods[_("User")])
         self.frame.UserSearchCombo.get_child().set_text(self.user)
-        self.frame.change_main_page(None, "search")
+        self.frame.change_main_page("search")
 
     def on_send_message(self, widget):
-        self.frame.privatechats.send_message(self.user, None, 1)
-        self.frame.change_main_page(None, "private")
+        self.frame.privatechats.send_message(self.user, show_user=True)
+        self.frame.change_main_page("private")
 
     def on_show_ip_address(self, widget):
 
-        if self.user not in self.frame.np.ip_requested:
-            self.frame.np.ip_requested.append(self.user)
-
+        self.frame.np.ip_requested.add(self.user)
         self.frame.np.queue.put(slskmessages.GetPeerAddress(self.user))
 
     def on_get_user_info(self, widget):
@@ -1172,7 +1229,7 @@ class PopupMenu(Gtk.Menu):
     def on_copy_user(self, widget):
         self.frame.clip.set_text(self.user, -1)
 
-    def on_give_privileges(self, widget):
+    def on_give_privileges(self, widget, error=None):
 
         self.frame.np.queue.put(slskmessages.CheckPrivileges())
 
@@ -1181,18 +1238,24 @@ class PopupMenu(Gtk.Menu):
         else:
             days = self.frame.np.privileges_left // 60 // 60 // 24
 
-        text = entry_dialog(
+        message = _("Give how many days of global privileges to this user?") + " (" + _("%(days)s days left") % {'days': days} + ")"
+
+        if error:
+            message += "\n\n" + error
+
+        days = entry_dialog(
             self.frame.MainWindow,
             _("Give privileges") + " " + _("to %(user)s") % {"user": self.user},
-            _("Give how many days of global privileges to this user?") + " (" + _("%(days)s days left") % {'days': days} + ")"
+            message
         )
 
-        if text:
+        if days:
             try:
-                days = int(text)
+                days = int(days)
                 self.frame.np.queue.put(slskmessages.GivePrivileges(self.user, days))
-            except Exception as e:
-                log.add_warning("%s", e)
+
+            except ValueError:
+                self.on_give_privileges(widget, error=_("Please enter a whole number!"))
 
     def on_private_rooms(self, widget, popup):
 
@@ -1243,64 +1306,55 @@ class TextSearchBar:
 
         self.textview.connect("key-press-event", self.on_key_press)
 
-    def on_search_changed(self, widget):
+    def on_search_match(self, search_type, restarted=False):
 
         buffer = self.textview.get_buffer()
-        start, end = buffer.get_bounds()
         query = self.entry.get_text()
 
         self.textview.emit("select-all", False)
 
-        iterator = start
-        match = iterator.forward_search(query, Gtk.TextSearchFlags.TEXT_ONLY | Gtk.TextSearchFlags.CASE_INSENSITIVE, limit=None)
+        if search_type == "typing":
+            start, end = buffer.get_bounds()
+            iterator = start
+        else:
+            current = buffer.get_mark("insert")
+            iterator = buffer.get_iter_at_mark(current)
+
+        if search_type == "previous":
+            match = iterator.backward_search(query, Gtk.TextSearchFlags.TEXT_ONLY | Gtk.TextSearchFlags.CASE_INSENSITIVE, limit=None)
+        else:
+            match = iterator.forward_search(query, Gtk.TextSearchFlags.TEXT_ONLY | Gtk.TextSearchFlags.CASE_INSENSITIVE, limit=None)
 
         if match is not None and len(match) == 2:
             match_start, match_end = match
-            buffer.place_cursor(match_end)
-            buffer.select_range(match_end, match_start)
+
+            if search_type == "previous":
+                buffer.place_cursor(match_start)
+                buffer.select_range(match_start, match_end)
+            else:
+                buffer.place_cursor(match_end)
+                buffer.select_range(match_end, match_start)
+
             self.textview.scroll_to_iter(match_start, 0, False, 0.5, 0.5)
 
+        elif not restarted and search_type != "typing":
+            start, end = buffer.get_bounds()
+
+            if search_type == "previous":
+                buffer.place_cursor(end)
+            elif search_type == "next":
+                buffer.place_cursor(start)
+
+            self.on_search_match(search_type, restarted=True)
+
+    def on_search_changed(self, widget):
+        self.on_search_match(search_type="typing")
+
     def on_search_previous_match(self, widget):
-
-        buffer = self.textview.get_buffer()
-        query = self.entry.get_text()
-
-        self.textview.emit("select-all", False)
-
-        current = buffer.get_mark("insert")
-        iterator = buffer.get_iter_at_mark(current)
-        match = iterator.backward_search(query, Gtk.TextSearchFlags.TEXT_ONLY | Gtk.TextSearchFlags.CASE_INSENSITIVE, limit=None)
-
-        if match is not None and len(match) == 2:
-            match_start, match_end = match
-            buffer.place_cursor(match_start)
-            buffer.select_range(match_start, match_end)
-            self.textview.scroll_to_iter(match_start, 0, False, 0.5, 0.5)
-        else:
-            start, end = buffer.get_bounds()
-            buffer.place_cursor(end)
-            self.on_search_previous_match(widget)
+        self.on_search_match(search_type="previous")
 
     def on_search_next_match(self, widget):
-
-        buffer = self.textview.get_buffer()
-        query = self.entry.get_text()
-
-        self.textview.emit("select-all", False)
-
-        current = buffer.get_mark("insert")
-        iterator = buffer.get_iter_at_mark(current)
-        match = iterator.forward_search(query, Gtk.TextSearchFlags.TEXT_ONLY | Gtk.TextSearchFlags.CASE_INSENSITIVE, limit=None)
-
-        if match is not None and len(match) == 2:
-            match_start, match_end = match
-            buffer.place_cursor(match_end)
-            buffer.select_range(match_end, match_start)
-            self.textview.scroll_to_iter(match_start, 0, False, 0.5, 0.5)
-        else:
-            start, end = buffer.get_bounds()
-            buffer.place_cursor(start)
-            self.on_search_next_match(widget)
+        self.on_search_match(search_type="next")
 
     def on_key_press(self, widget, event):
         key = Gdk.keyval_name(event.keyval)
@@ -1347,21 +1401,29 @@ def human_speed(filesize):
 
 
 def humanize(number):
-    fashion = DECIMALSEP
+
+    fashion = NICOTINE.np.config.sections["ui"]["decimalsep"]
+
     if fashion == "" or fashion == "<None>":
         return str(number)
+
     elif fashion == "<space>":
         fashion = " "
+
     number = str(number)
+
     if number[0] == "-":
         neg = "-"
         number = number[1:]
     else:
         neg = ""
+
     ret = ""
+
     while number[-3:]:
         part, number = number[-3:], number[:-3]
         ret = "%s%s%s" % (part, fashion, ret)
+
     return neg + ret[:-1]
 
 
@@ -1444,3 +1506,100 @@ def _expand_alias(aliases, cmd):
         log.add_warning("%s", error)
 
     return ""
+
+
+""" Fonts and Colors """
+
+
+def parse_color_string(color_string):
+    """ Take a color string, e.g. BLUE, and return a HEX color code """
+
+    if color_string:
+        color_rgba = Gdk.RGBA()
+
+        if color_rgba.parse(color_string):
+            color_hex = "#%02X%02X%02X" % (round(color_rgba.red * 255), round(color_rgba.green * 255), round(color_rgba.blue * 255))
+            return color_hex
+
+    return None
+
+
+def set_list_color(listview, color):
+
+    for c in listview.get_columns():
+        for r in c.get_cells():
+            if isinstance(r, (Gtk.CellRendererText, Gtk.CellRendererCombo)):
+                set_widget_color(r, color)
+
+
+def set_list_font(listview, font):
+
+    for c in listview.get_columns():
+        for r in c.get_cells():
+            if isinstance(r, (Gtk.CellRendererText, Gtk.CellRendererCombo)):
+                set_widget_font(r, font)
+
+
+def set_widget_color(widget, color):
+
+    if color:
+        widget.set_property("foreground", color)
+    else:
+        widget.set_property("foreground-set", False)
+
+
+css_providers = {}
+
+
+def set_widget_fg_bg_css(widget, bg_color=None, fg_color=None):
+
+    class_name = "widget_custom_color"
+    css = "." + class_name + " {"
+
+    bg_color_hex = parse_color_string(bg_color)
+    fg_color_hex = parse_color_string(fg_color)
+
+    if bg_color_hex is not None:
+        css += "background: " + bg_color_hex + ";"
+
+    if fg_color_hex is not None:
+        css += "color: " + fg_color_hex + ";"
+
+    css += "}"
+
+    context = widget.get_style_context()
+
+    if widget not in css_providers:
+        css_providers[widget] = Gtk.CssProvider()
+        context.add_provider(css_providers[widget], Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
+        context.add_class(class_name)
+
+    css_providers[widget].load_from_data(css.encode('utf-8'))
+
+
+def set_widget_font(widget, font):
+    widget.set_property("font", font)
+
+
+def update_widget_visuals(widget, list_font_target="listfont", update_text_tags=True):
+
+    if isinstance(widget, Gtk.ComboBox) and widget.get_has_entry() or \
+            isinstance(widget, Gtk.Entry):
+        if isinstance(widget, Gtk.ComboBox):
+            widget = widget.get_child()
+
+        set_widget_fg_bg_css(
+            widget,
+            bg_color=NICOTINE.np.config.sections["ui"]["textbg"],
+            fg_color=NICOTINE.np.config.sections["ui"]["inputcolor"]
+        )
+
+    elif update_text_tags and isinstance(widget, Gtk.TextTag):
+        # Chat rooms and private chats have their own code for this
+
+        set_widget_color(widget, NICOTINE.np.config.sections["ui"]["chatremote"])
+        set_widget_font(widget, NICOTINE.np.config.sections["ui"]["chatfont"])
+
+    elif isinstance(widget, Gtk.TreeView):
+        set_list_color(widget, NICOTINE.np.config.sections["ui"]["search"])
+        set_list_font(widget, NICOTINE.np.config.sections["ui"][list_font_target])
