@@ -20,6 +20,7 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+import importlib
 import os
 import pickle
 import re
@@ -27,16 +28,25 @@ import shelve
 import stat
 import string
 import sys
-import taglib
 import _thread
 
 from gettext import gettext as _
 
 from pynicotine import slskmessages
 from pynicotine.logfacility import log
+from pynicotine.metadata.tinytag import TinyTag
 
-if sys.platform == "win32":
-    # Use semidbm for faster shelves on Windows
+""" Check if there's an appropriate (performant) database type for shelves """
+
+if importlib.util.find_spec("_gdbm"):
+
+    def shelve_open_gdbm(filename, flag='c', protocol=None, writeback=False):
+        import _gdbm
+        return shelve.Shelf(_gdbm.open(filename, flag), protocol, writeback)
+
+    shelve.open = shelve_open_gdbm
+
+elif importlib.util.find_spec("semidbm"):
 
     def shelve_open_semidbm(filename, flag='c', protocol=None, writeback=False):
         import semidbm
@@ -44,15 +54,24 @@ if sys.platform == "win32":
 
     shelve.open = shelve_open_semidbm
 
+else:
+    print(_("Cannot find %(option1)s or %(option2)s, please install either one.") % {
+        "option1": "gdbm",
+        "option2": "semidbm"
+    })
+    sys.exit()
+
 
 class Shares:
 
-    def __init__(self, np, config, queue, ui_callback=None):
+    def __init__(self, np, config, queue, ui_callback=None, connected=False):
         self.np = np
         self.ui_callback = ui_callback
         self.config = config
         self.queue = queue
+        self.connected = connected
         self.translatepunctuation = str.maketrans(dict.fromkeys(string.punctuation, ' '))
+        self.tinytag = TinyTag()
 
         self.convert_shares()
         self.load_shares(
@@ -70,42 +89,48 @@ class Shares:
             ]
         )
 
-        self.compressed_shares_buddy = self.compressed_shares_normal = None
-
-        if not self.config.sections["transfers"]["friendsonly"]:
-            self.compress_shares("normal")
-
-        if self.config.sections["transfers"]["enablebuddyshares"]:
-            self.compress_shares("buddy")
+        self.create_compressed_shares_message("normal")
+        self.create_compressed_shares_message("buddy")
 
         self.newbuddyshares = self.newnormalshares = False
 
+    def set_connected(self, connected):
+        self.connected = connected
+
     """ Shares-related actions """
 
     def real2virtual(self, path):
         path = os.path.normpath(path)
 
         for (virtual, real) in self._virtualmapping():
+            # Remove slashes from share name to avoid path conflicts
+            virtual = virtual.replace('/', '_').replace('\\', '_')
+
             real = os.path.normpath(real)
 
             if path == real:
                 return virtual
+
             if path.startswith(real + os.sep):
                 virtualpath = virtual + '\\' + path[len(real + os.sep):].replace(os.sep, '\\')
                 return virtualpath
+
         return "__INTERNAL_ERROR__" + path
 
     def virtual2real(self, path):
         path = os.path.normpath(path)
 
         for (virtual, real) in self._virtualmapping():
-            virtual = os.path.normpath(virtual)
+            # Remove slashes from share name to avoid path conflicts
+            virtual = os.path.normpath(virtual.replace('/', '_').replace('\\', '_'))
 
             if path == virtual:
                 return real
+
             if path.startswith(virtual + '\\'):
                 realpath = real + path[len(virtual):].replace('\\', os.sep)
                 return realpath
+
         return "__INTERNAL_ERROR__" + path
 
     def _virtualmapping(self):
@@ -141,10 +166,14 @@ class Shares:
             try:
                 self.config.sections["transfers"][destination] = shelve.open(shelvefile, protocol=pickle.HIGHEST_PROTOCOL)
             except Exception:
+                from traceback import format_exc
+
                 errors.append(shelvefile)
+                exception = format_exc()
 
         if errors:
             log.add_warning(_("Failed to process the following databases: %(names)s") % {'names': '\n'.join(errors)})
+            log.add_warning(exception)
 
             self.clear_shares()
 
@@ -152,6 +181,8 @@ class Shares:
 
     def set_shares(self, sharestype="normal", files=None, streams=None, mtimes=None, wordindex=None, fileindex=None):
 
+        self.config.create_data_folder()
+
         if sharestype == "normal":
             storable_objects = [
                 (files, "sharedfiles", "files.db"),
@@ -185,24 +216,32 @@ class Shares:
         self.set_shares(sharestype="normal", files={}, streams={}, mtimes={}, wordindex={}, fileindex={})
         self.set_shares(sharestype="buddy", files={}, streams={}, mtimes={}, wordindex={}, fileindex={})
 
+    def create_compressed_shares_message(self, sharestype):
+
+        """ Create a message that will later contain a compressed list of our shares """
+
+        if sharestype == "normal":
+            self.compressed_shares_normal = slskmessages.SharedFileList(
+                None,
+                self.config.sections["transfers"]["sharedfilesstreams"]
+            )
+
+        elif sharestype == "buddy":
+            self.compressed_shares_buddy = slskmessages.SharedFileList(
+                None,
+                self.config.sections["transfers"]["bsharedfilesstreams"]
+            )
+
     def compress_shares(self, sharestype):
 
-        if sharestype == "normal":
-            streams = self.config.sections["transfers"]["sharedfilesstreams"]
-        elif sharestype == "buddy":
-            streams = self.config.sections["transfers"]["bsharedfilesstreams"]
-
-        if streams is None:
-            log.add_warning(_("ERROR: No %(type)s shares database available") % {"type": sharestype})
-            return
-
-        m = slskmessages.SharedFileList(None, streams)
-        _thread.start_new_thread(m.make_network_message, (0, True))
+        """ Begin compressing the shares list. This compressed list will be used to
+        quickly send our file list to users. """
 
         if sharestype == "normal":
-            self.compressed_shares_normal = m
+            _thread.start_new_thread(self.compressed_shares_normal.make_network_message, (0, True))
+
         elif sharestype == "buddy":
-            self.compressed_shares_buddy = m
+            _thread.start_new_thread(self.compressed_shares_buddy.make_network_message, (0, True))
 
     def close_shares(self):
         for db in [
@@ -295,13 +334,25 @@ class Shares:
             if self.ui_callback:
                 self.ui_callback.rescan_finished(sharestype)
 
+            self.create_compressed_shares_message(sharestype)
             self.compress_shares(sharestype)
-            self.send_num_shared_folders_files()
 
-        except Exception as ex:
+            if self.connected:
+                """ Don't attempt to send file stats to the server before we're connected. If we skip the
+                step here, it will be done once we log in instead ("login" function in pynicotine.py). """
+
+                self.send_num_shared_folders_files()
+
+        except Exception:
+            from traceback import format_exc
+
             log.add(
-                _("Failed to rebuild share, serious error occurred. If this problem persists delete %s/*.db and try again. If that doesn't help please file a bug report with the stack trace included (see terminal output after this message). Technical details: %s"), (self.config.data_dir, ex)
+                _("Serious error occurred while rescanning shares. If this problem persists, delete %(dir)s/*.db and try again. If that doesn't help, please file a bug report with this stack trace included: %(trace)s"), {
+                    "dir": self.config.data_dir,
+                    "trace": "\n" + format_exc()
+                }
             )
+
             if self.ui_callback:
                 self.ui_callback.hide_scan_progress(sharestype)
 
@@ -418,7 +469,7 @@ class Shares:
 
         shared[vdir] = shared.get(vdir, [])
 
-        if file not in [i[0] for i in shared[vdir]]:
+        if file not in (i[0] for i in shared[vdir]):
             fileinfo = self.get_file_info(file, name)
             shared[vdir] += [fileinfo]
 
@@ -461,7 +512,7 @@ class Shares:
 
         bshared[vdir] = bshared.get(vdir, [])
 
-        if file not in [i[0] for i in bshared[vdir]]:
+        if file not in (i[0] for i in bshared[vdir]):
 
             fileinfo = self.get_file_info(file, name)
             bshared[vdir] += [fileinfo]
@@ -486,20 +537,20 @@ class Shares:
         try:
             for entry in os.scandir(folder):
                 if entry.is_dir():
-
-                    path = entry.path
+                    path = entry.path.replace('\\', os.sep)
 
                     if self.is_hidden(path):
-                        return mtimes
+                        continue
 
                     try:
                         mtime = entry.stat().st_mtime
+
                     except OSError as errtuple:
                         log.add(_("Error while scanning %(path)s: %(error)s"), {
                             'path': path,
                             'error': errtuple
                         })
-                        return mtimes
+                        continue
 
                     mtimes[path] = mtime
                     dircontents = self.get_folder_mtimes(path)
@@ -575,10 +626,12 @@ class Shares:
         return files, streams
 
     def get_file_info(self, name, pathname, file=None):
-        """ Get metadata via taglib """
+        """ Get file metadata """
 
         try:
             audio = None
+            bitrateinfo = None
+            duration = None
 
             if file:
                 # Faster way if we use scandir
@@ -586,23 +639,27 @@ class Shares:
             else:
                 size = os.stat(pathname).st_size
 
+            """ We skip metadata scanning of files without meaningful content """
             if size > 128:
-                """ On some operating systems, TagLib might crash when scanning files
-                with truncated headers. We skip metadata scanning of files without
-                meaningful content to work around this issue. """
-
                 try:
-                    audio = taglib.File(pathname)
-                except IOError:
-                    pass
+                    audio = self.tinytag.get(pathname, size, tags=False)
+
+                except Exception as errtuple:
+                    log.add(
+                        _("Error while scanning metadata for file %(path)s: %(error)s"), {
+                            'path': pathname,
+                            'error': errtuple
+                        }
+                    )
 
             if audio is not None:
-                bitrateinfo = (int(audio.bitrate), int(False))  # Second argument used to be VBR (variable bitrate)
-                fileinfo = (name, size, bitrateinfo, int(audio.length))
-            else:
-                fileinfo = (name, size, None, None)
+                if audio.bitrate is not None:
+                    bitrateinfo = (int(audio.bitrate), int(False))  # Second argument used to be VBR (variable bitrate)
 
-            return fileinfo
+                if audio.duration is not None:
+                    duration = int(audio.duration)
+
+            return (name, size, bitrateinfo, duration)
 
         except Exception as errtuple:
             log.add(_("Error while scanning file %(path)s: %(error)s"), {'path': pathname, 'error': errtuple})
@@ -619,26 +676,34 @@ class Shares:
             stream.extend(message.pack_object(fileinfo[0]))
             stream.extend(message.pack_object(fileinfo[1], unsignedlonglong=True))
 
-            if fileinfo[2] is not None:
-                try:
-                    stream.extend(message.pack_object('mp3'))
-                    stream.extend(message.pack_object(3))
+            if fileinfo[2] is not None and fileinfo[3] is not None:
+                stream.extend(message.pack_object('mp3'))
+                stream.extend(message.pack_object(3))
 
-                    stream.extend(message.pack_object(0))
+                stream.extend(message.pack_object(0))
+                try:
                     stream.extend(message.pack_object(fileinfo[2][0], unsignedint=True))
-                    stream.extend(message.pack_object(1))
-                    stream.extend(message.pack_object(fileinfo[3], unsignedint=True))
-                    stream.extend(message.pack_object(2))
-                    stream.extend(message.pack_object(fileinfo[2][1]))
+
                 except Exception:
-                    log.add(_("Found meta data that couldn't be encoded, possible corrupt file: '%(file)s' has a bitrate of %(bitrate)s kbs, a length of %(length)s seconds and a VBR of %(vbr)s"), {
-                        'file': fileinfo[0],
-                        'bitrate': fileinfo[2][0],
-                        'length': fileinfo[3],
-                        'vbr': fileinfo[2][1]
-                    })
-                    stream.extend(message.pack_object(''))
+                    # Invalid bitrate
                     stream.extend(message.pack_object(0))
+
+                stream.extend(message.pack_object(1))
+                try:
+                    stream.extend(message.pack_object(fileinfo[3], unsignedint=True))
+
+                except Exception:
+                    # Invalid length
+                    stream.extend(message.pack_object(0))
+
+                stream.extend(message.pack_object(2))
+                try:
+                    stream.extend(message.pack_object(fileinfo[2][1]))
+
+                except Exception:
+                    # Invalid VBR value
+                    stream.extend(message.pack_object(0))
+
             else:
                 stream.extend(message.pack_object(''))
                 stream.extend(message.pack_object(0))
@@ -724,7 +789,7 @@ class Shares:
             # DB is closed, perhaps when rescanning share or closing Nicotine+
             return
 
-    def process_search_request(self, searchterm, user, searchid, direct=0):
+    def process_search_request(self, searchterm, user, searchid, direct=False):
         """ Note: since this section is accessed every time a search request arrives,
         several times a second, please keep it as optimized and memory
         sparse as possible! """
@@ -736,8 +801,9 @@ class Shares:
         if searchterm is None:
             return
 
-        if user == self.config.sections["server"]["login"]:
-            # We shouldn't send a search response if we initiated the search request
+        if not direct and user == self.config.sections["server"]["login"]:
+            # We shouldn't send a search response if we initiated the search request,
+            # unless we're specifically searching our own username
             return
 
         maxresults = self.config.sections["searches"]["maxresults"]
@@ -775,11 +841,6 @@ class Shares:
             queuesizes = self.np.transfers.get_upload_queue_sizes()
             slotsavail = self.np.transfers.allow_new_uploads()
 
-            if reason == "geoip":
-                geoip = 1
-            else:
-                geoip = 0
-
             if checkuser == 2:
                 fileindex = self.config.sections["transfers"]["bfileindex"]
             else:
@@ -790,11 +851,11 @@ class Shares:
             message = slskmessages.FileSearchResult(
                 None,
                 self.config.sections["server"]["login"],
-                geoip, searchid, resultlist, fileindex, slotsavail,
+                searchid, resultlist, fileindex, slotsavail,
                 self.np.speed, queuesizes, fifoqueue, numresults
             )
 
-            self.np.process_request_to_peer(user, message)
+            self.np.send_message_to_peer(user, message)
 
             if direct:
                 log.add_search(
